/*  $Id: pion_server.pl,v 2.1.0 2009/4/27 13:50:30 huang Exp $


    Author:        Zhisheng Huang
    E-mail:        huang@cs.vu.nl
    WWW:           http://www.cs.vu.nl/~huang
    Copyright (C): 2004-2009 Vrije University Amsterdam

This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    As a special exception, if you link this library with other files,
    compiled with a Free Software compiler, to produce an executable, this
    library does not by itself cause the resulting executable to be covered
    by the GNU General Public License. This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.

*/

% this server requires XDIG version 2.1.0 or higher.

user:file_search_path(xdig, '../xdig').


:-use_module(xdig('dig_client')).
:-use_module(xdig('dig_server')).
:-use_module(xdig('dig_process')).
:-use_module(xdig('owl/owl2dig')).
:-use_module(xdig('dig_db')).

:-use_module(library('http/thread_httpd')).
:-use_module(library('http/http_open')).

:-use_module(xdig('pion/pion_process')).
:-use_module(xdig('pion/pion_selection')).

:- use_module('http_util').
	
:-dynamic
	ngd/2.

:-multifile
      selected_set/5,
       pion_selection_function/10.
 


%deal with the tell request
my_dig_server_processing(RequestData, Answer, [connection(close)]):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type=tells,
			!,
                  pion_setting(kb, KBID),
                  dig_assert_data(KBID, Data),
                  dig_db_post(KBID, withClearKB, Answer, [connection(close)]),
 %                 RequestData= elements([element(tells, XMLNS,_)|_]),% for namespace processing
 %                 namespace_processing(XMLNS), 
                  true.



%deal with the ask request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type=asks,
			!,  
%                  format(user_error, 'RequestData=~q~n', [RequestData]),
                  RequestData= elements([element(asks, Atts, Data)]),
                  dig_kburi(default, KBURI),
                  append(Atts, [uri=KBURI], AttswithURI),
                  RequestDatawithURI= elements([element(asks, AttswithURI, Data)]),           
			dig_post(RequestDatawithURI, Answer1, [connection(close)]),
%                  format(user_error, 'Answer1=~q~n~n~n', [Answer1]),
			pion_answer_preprocessing(Answer1, Data, Answer),
                  true.


%deal with owl tell request
my_dig_server_processing(RequestData, Answer, [connection(close)]):-
			dig_requestdata_analysis(RequestData, _Data, Type),
			Type = 'rdf:RDF',
                  !,
               	owl2dig(RequestData, elements(L)),
                  dig_get_element(L, E, _),
                  E = element(tells, _, L1),
                 	pion_setting(kb, KBID),
 		     	dig_assert_data(KBID, L1),
                  dig_db_post(KBID, withClearKB, Answer, [connection(close)]),	
			true.


%deal with the set request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = set,
			!,
			dig_setting_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_, Reply),
                  true.

%deal with the get request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = get,
			!,
			pion_getting_processing(Data, Reply),
                  Answer = answer(_, text(Reply)),
                  true.


%deal with the do operation
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = do,
			!,
			dig_do_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_, Reply),
                  true.

%deal with the newKB request
my_dig_server_processing(RequestData, Answer, [connection(close)]):-
			dig_requestdata_analysis(RequestData, _Data, Type),
			Type = newKB,
			!,
                 dig_db_element(setting(kburi, default), URI),
%                 pion_setting(kb, KBID),//to be checked
                  sformat(Reply, '<?xml version="1.0" encoding="UTF-8"?><response><kb uri=~q/></response>\n',[URI]),
                  Answer = answer(_, text(Reply)),
                  true.





%deal with other request
 my_dig_server_processing(RequestData, Answer, Options):-
			dig_post(RequestData, Answer, Options).


%pion_answer_preprocessing(+Answer, +Queries, -NewAnswer):-
%working on the Answer from the server with respect to the Queries
%if the current KB is inconsistent, then do inconsistency processing
%to obtain the NewAnswer, otherwise use the current Answer.


pion_answer_preprocessing(Answer1, Queries, Answer):-
    Answer1 = answer(_Header, AnswerBody),
    (AnswerBody = [element(responses, _, Elements)|_];AnswerBody = [element(response, _, Elements)|_]),
    pion_setting(kb, KBID),
    (pattern_set(Elements, incoherent); (dig_consistent_check(kb(KBID), Status),Status=false)),
    !,
    pion_setting(strategy, Strategy),
    pion_inconsistency_processing(KBID, Queries, Elements, Strategy, Answer).

pion_answer_preprocessing(Answer1, _Queries, Answer):-
		Answer = Answer1.

  			
%pion_inconsistency_processing(+KBID, +Queries, +AnswerFromServer, +Strategy, -MyAnswer)

pion_inconsistency_processing(KBID, Queries, _As, Strategy, Answer):- 
    pion_separation(Queries, Qs1, Qs2),
    pion_queries_answers(KBID, Qs1, Strategy, As1),
    dig_db_post(KBID, withClearKB, _, [connection(close)]),
    dig_ask(elements(Qs2), Answer1, [connection(close)]),
    dig_get_response_data(Answer1, As2),
%%  format(user_error, 'As1:~w~n As2:~w~n', [As1, As2]),
    append(As1, As2, As),
    !,
    dig_response_elements(As, Answer),
    dig_db_post(KBID, withClearKB, _, [connection(close)]).


pion_separation([], [], []):-!. 

pion_separation([H|T], [H|T1], T2):-
     H = element(Type, _, _),
     (Type = instance; Type = subsumes),
     !,
     pion_separation(T, T1, T2).

pion_separation([H|T], T1, [H|T2]):-!,
     pion_separation(T, T1, T2).





pion_queries_answers(_KBID, [], _, []):-!.

pion_queries_answers(KBID, [H|T], Strategy,  [A|As]):-
%   format(user_error, 'processing ~q~n~n', [H]),
    pion_query_answer(KBID, H, Strategy, A),
%    format(user_error, 'get the answer ~q~n~n', [A]),
    !,
    pion_queries_answers(KBID, T, Strategy, As).




pion_query_answer(KBID, Q, Strategy, AnswerElement):-
       Q = element(instance, X, L),
       !,
      dig_get_element(L, element(individual, [name=I], []), L1),
      dig_get_element(L1, E, _),
      dig_get_parentconcept(KBID, I, L2),
      Q1 = element(subsumes, X, [E,L2]),   
	pion_query_answer(KBID, Q1, Strategy, AnswerElement),
      true.


pion_query_answer(KBID, Q, Strategy, A):-
     Q = element(subsumes, [], _L),      
      !,
	query_processing(Q, nil, KBID, Strategy, A),
      true.

pion_query_answer(KBID, Q, Strategy, A):-
     Q = element(subsumes, [id=ID], _L), 
      !,
	query_processing(Q, ID, KBID, Strategy, A), 
      true.
     
pion_query_answer(_KBID, Q, _, A):-
       Q = element(_, _, _),
       !,
       A = element(error, [unsupportedQueryFormat = Q],[]).



dig_get_parentconcept(KBID, I, E):-
   dig_db_element(KBID, element(instanceof, _, L)),
   dig_get_element(L, element(individual, [name=I],[]), L2),
   dig_get_element(L2, E, _),
   \+ E = element(top, [],[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% the start of  the linear extension strategy, 
%%%  version 1.2.4 renewed
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


query_processing(Q, ID, KBID, Strategy, A):-
          Strategy = linear_extension,
          !,
          pion_setting(selectionFunction, Type), 
	    linear_query_processing(Type, Q, ID, 1, [], icr([],[],[]), KBID, A).

query_processing(Q, ID, KBID, Strategy, A):-
          Strategy = k_extension,
          !,
          dig_db_element(setting(pion, kvalue), K1),
          term_to_atom(K, K1),
          pion_setting(selectionFunction, Type), 
	    k_query_processing(K, Type, Q, ID, 1, [], icr([],[],[]), KBID, A).



query_processing(_Q, _ID, _KBID, Strategy, A):-
        A = element(error, [unsupportedStrategy = Strategy],[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% k extension proceesing
%  version 1.5 added
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%k_query_processing/9
% starting liner query processing for selection function Type on 
% query Q with ID at Step K1
% the current working set is P, its ICR is ICRp
% on knowledge bases KBID
% return the Answer A.

k_query_processing(K, Type, Q, ID, K1, P, ICRp, KBID, A):-
     pion_selection_function(Type, KBID, Q, K1, P, ICRp, N, L, ICR, Status),
     !,
%     format(user_error, 'K=~q, Q=~q, P=~q~n,Status=~q, N=~q~n', [K, Q, P, Status, N]),
     k_query_processing(K, Type, Q, ID, K1, P, KBID, N, L, ICR, Status, A).


%k_query_processing/12
% k query processing for selection function Type on 
% query Q with ID at Step K1
% the current working set is P, its ICR is ICRp
% on knowledge bases KBID
% after the selection function have selected a new selected set N,
% the new working set is L with its ICR set ICR for which the consistency status
% is Status1,
% return the Answer A.

% if there is no new selected set and the current working set is consistent, then take the current answer.
k_query_processing(_Kc, _Type, Q, ID, K1, _P, _KBID, [], L, _ICR, S, A):-
     S= true,
     !,
     query_post(L, Q, Status2),    
     A = element(Status2, [id=ID, k=K1+0],[]).

% if there is no new selected set and the current working set is inconsistent, then do overdetermined processing.
k_query_processing(_Kc, _Type, Q, ID, K1, _P, _KBID, [], L, _ICR, _S, A):-
%     format(user_error, 'S=~w~n', [S]),
     !,
     dig_db_element(setting(pion, odp),  Strategy),
     K2 is K1+1,

     overdetermined_processing(Strategy, [[], Q, L], [Status2, L2, K3]),
% where we use the whole working set L to remove some axioms from it 
% in the overdetermined processing

     explain_answer(Status2, ID, L2, K2, K3, A).




% if Kc=1 and the current working set is inconsistent, then do overdetermined processing
k_query_processing(Kc, _Type, Q, ID, K1, _P, _KBID, _N, L, _ICR, S, A):-
     Kc=1,
     S= false,
     !,
     dig_db_element(setting(pion, odp),  Strategy),
     K2 is K1+1,
 %    format(user_error, '***S***=~w~nP=~w~nN=~w~nL=~w~n', [S,P,N,L]),

     overdetermined_processing(Strategy, [[], Q, L], [Status2, L2, K3]),
% where the previous selected consistent set is set to be [], use the whole working set L
% to remove some axioms from it. The list N consists of only currently selected new set at step K1, and
% P is the previous selected consistent set at step K1-Ks. Thus, P\cup N =\= L

     explain_answer(Status2, ID, L2, K2, K3, A).




% if Kc=1 and the current working set is consistent, then  do k query processing with as the original k value
k_query_processing(Kc, Type, Q, ID, K1, _P, KBID, _N, L, ICR, S, A):-
     Kc=1,
     S = true,
     !,
     query_post(L, Q, Status2), 
     K2 is K1+1,
     dig_db_element(setting(pion, strategy), k_extension),
     dig_db_element(setting(pion, kvalue), Kss),
     term_to_atom(Ks, Kss),
     k_query_processing2(Ks, Type, Q, ID, K2, KBID, L, ICR, Status2, A).




% if Kc > 1 then  do k query processing with Kc:=Kc-1
k_query_processing(Kc, Type, Q, ID, K1, P, KBID, N, _L, ICR, _S, A):-
     Kc >1,
     !,
     K2 is K1+1,
     K3 is Kc-1,   
     pion_selection_function(Type, KBID, Q, K2, N, ICR, N1, L1, ICR2, Status),
     k_query_processing(K3, Type, Q, ID, K2, P, KBID, N1, L1, ICR2, Status, A),
     true.




 
% otherwise return error  
k_query_processing(_K, _Type, _Q, ID, _K1, _P,  _KBID, _N, _L, _ICR, _Status, A):-
     A = element(error, [id=ID], []).


%k_query_processing2/10
% if the answer is undetermined, then continue to extend the working set
k_query_processing2(_Kc, Type, Q, ID, K1, KBID, L, ICR, Status2, A):-
      Status2 = undetermined,     
      !,
      K2 is K1+1,
%      dig_db_element(setting(pion, strategy), k_extension(Ks)),
     dig_db_element(setting(pion, strategy), k_extension),
     dig_db_element(setting(pion, kvalue), Kss),
     term_to_atom(Ks, Kss),
      k_query_processing(Ks, Type, Q, ID, K2, L, ICR, KBID, A).

% if the answer is certain, then stop.
k_query_processing2(_Kc, _Type, _Q, ID, K, _KBID, L, _ICR, Status2, A):-
(Status2 = accepted; Status2=rejected),
     !,
     explain_answer(Status2, ID, L, K, 0,  A).

% show unknown status.
k_query_processing2(_K, _Type, Q, ID, K1, _KBID, _L, _ICR, Status2, A):-
      format(user_error, 'do not know how to deal it:~w, Status2:~w~n', [Q, Status2]),
      A = element(error, [id=ID, k= K1, message='cannot deal with the query.'],[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% the end of  the k extension strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%linear_query_processing/8
% starting liner query processing for selection function Type on 
% query Q with ID at Step K
% the current working set is P, its ICR is ICRp
% on knowledge bases KBID
% return the Answer A.

linear_query_processing(Type, Q, ID, K, P, ICRp, KBID, A):-
     pion_selection_function(Type, KBID, Q, K, P, ICRp, N, L, ICR, Status),
     !,
%     format(user_error, 'K=~q, Q=~q, P=~q~n,Status=~q, N=~q~n', [K, Q, P, Status, N]),
     linear_query_processing(Type, Q, ID, K, P, KBID, N, L, ICR, Status, A).


%linear_query_processing/11
% liner query processing for selection function Type on 
% query Q with ID at Step K
% the current working set is P, its ICR is ICRp
% on knowledge bases KBID
% after the selection function have selected a new selected set N,
% the new working set is L with its ICR set ICR for which the consistency status
% is Status1,
% return the Answer A.

% if there is no new selected set, then take the current answer.
linear_query_processing(_Type, Q, ID, K, _P, _KBID, [], L, _ICR, _Status1, A):-
     !,
     query_post(L, Q, Status2), 
     A = element(Status2, [id=ID, k=K+0],[]).


% if the current working set is inconsistent, then do overdetermined processing
linear_query_processing(_Type, Q, ID, K, P, _KBID, N, _L, _ICR, false, A):-
     !,
     dig_db_element(setting(pion, odp),  Strategy),
     K1 is K+1,
     overdetermined_processing(Strategy, [P, Q, N], [Status2, L2, K2]),
     explain_answer(Status2, ID, L2, K1, K2, A).




% if the current working set is consistent, then  continue to do linear query processing k+1
linear_query_processing(Type, Q, ID, K, _P, KBID, _N, L, ICR, true, A):-
     !,
     query_post(L, Q, Status2), 
     K1 is K+1,
     linear_query_processing2(Type, Q, ID, K1, KBID, L, ICR, Status2, A).


 
% otherwise return error  
linear_query_processing(_Type, _Q, ID, _K, _P,  _KBID, _N, _L, _ICR, _Status, A):-
     A = element(error, [id=ID], []).


%linear_query_processing2/9

% if the answer is undetermined, then continue to extend the working set
linear_query_processing2(Type, Q, ID, K1, KBID, L, ICR, Status2, A):-
      Status2 = undetermined,
      !,
      linear_query_processing(Type, Q, ID, K1, L, ICR, KBID, A).

% if the answer is certain, then stop.
linear_query_processing2(_Type, _Q, ID, K, _KBID, L, _ICR, Status2, A):-
(Status2 = accepted; Status2=rejected),
     !,
     explain_answer(Status2, ID, L, K, 0,  A).


% show unknown status.
linear_query_processing2(_Type, Q, ID, K1, _KBID, _L, _ICR, Status2, A):-
      format(user_error, 'do not know how to deal it:~w, Status2:~w~n', [Q, Status2]),
      A = element(error, [id=ID, k= K1, message='cannot deal with the query.'],[]).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% the end of  the linear extension strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


overdetermined_processing(first_maximal_consistent_set, [P, Q, L], [Status1, L2, K]):-
   !,
   retract(pion_log_counter(odp,_)),
   assert(pion_log_counter(odp, 0)),
   overdetermined_processing(P, Q, L, Status1, L2, K).

overdetermined_processing(Strategy, [P, Q, L], [Status1, L2, K]):-
   Strategy = path_pruning_with_google_distance,
   !,
   retract(pion_log_counter(odp,_)),
   assert(pion_log_counter(odp, 0)),
   odp(Strategy, P, Q, L, Status1, L2, K),
   true.


% first_maximal_coherent_set as the default strategy.
 overdetermined_processing(_Strategy, [P, Q, L], [Status1, L2, K]):-
%  format(user_error, '***Q=~w~n', [Q]),
   overdetermined_processing(P, Q, L, Status1, L2, K).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% overdetermined processing
%%% first_maximal_consistent_set_strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%overdetermined_processing(+WorkingSet, +Query,  +CurrentSelectedSet, -Status, -NewWorkingSet, -KCounter)

overdetermined_processing(_L1, _Q, _L, Status, _L2, K):-
	pion_log_counter(odp, I),
      dig_db_element(setting(pion, odpLimit), IL),
      I > IL,
      !,
      Status = undetermined,
      K = IL.


% search for the maximal conherent set by deleting one element
overdetermined_processing(L1, Q, L, Status1, L2, K):-
            select(_E, L, H),
%          show_overdetermined_processing(E), % for debugging only
            union(L1, H, L2),
            query_post(L2, Q, Status),       
%            format(user_error, 'L2=~w~nQ=~w~nStatus=~w~n', [L2, Q, Status]),
     		pion_log_counter(odp, I),
            I1 is I+1,
            retract(pion_log_counter(odp, I)),
            assert(pion_log_counter(odp, I1)),
 %           format(user_error, 'Q:~w, Status:~w, ignoring ~w~n', [Q, Status, E]),
		(Status = accepted; Status = rejected; Status= undetermined),           
            !,
            Status1 = Status,
            K = I1,
            true.

% if deleting one element is not enough to get a consistent set,
% then deleting one more to continue the processing.

overdetermined_processing(L1, Q, L, Status1, L2, K1):-
            select(_E, L, H),
            H \= [],
            !,
            overdetermined_processing(L1, Q, H, Status, L2, K),
            K1 is K +1,
            Status1 = Status.



overdetermined_processing(_L1, _Q, _, Status, [element(message,['fail to find relevant consistent set'], [])], K):-
         !,
	   Status = undetermined,
         K =1.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% end of first_maximal_consistent_set_strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% overdetermined processing
%%% path pruning 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%odp(+Strategy, +PreSelected, +Query, +NewlySelected, -Status, -Explaination, -K).


   odp(path_pruning_with_google_distance, P, Q, L, Status1, L3, K):-
 %      format(user_error, 'Q=~w~nP=~w~nL=~w~n', [Q, P, L]),
       DistanceMeasure=google,
       get_query_distance(Q, DistanceMeasure, Distance, CoreKeyword),
       union(P, L, W),
       path_pruning(W, Distance, CoreKeyword, DistanceMeasure, L2),
       query_post(L2, Q, Status),       
%      format(user_error, '**W=~w~nL2=~w~nQ=~w~nStatus=~w~n~n', [W, L2, Q, Status]),
       pion_log_counter(odp, I),
       I1 is I+1,
       retract(pion_log_counter(odp, I)),
       assert(pion_log_counter(odp, I1)),
%     format(user_error, 'Q:~w, Status:~w~n', [Q, Status]),
      (Status = overdetermined -> overdetermined_processing([], Q, L2, Status1, L3, K);(Status1 = Status,K = I1, L3=L2)),
%     format(user_error, '*Q:~w, Status1:~w K:~w~nL3=~w~n', [Q, Status1,K, L3]),
       true.


get_query_distance(Q, google, Distance, CoreKeyword):-
      !,
      get_corekeyword(Q, CoreKeyword),
      pion_icr_set_element(Q, icr(_I, C, _R)),
      google_distance_element(google_distance_max, C, CoreKeyword, Distance),
      true.

get_query_distance(Q, DistanceMeasure, Distance, CoreKeyword):-
      get_corekeyword(Q, CoreKeyword),
      get_axiom_distance(CoreKeyword, Q, DistanceMeasure, Distance),
      true.

get_corekeyword(Q, CoreKeyword):-
     Q=element(_,_, L1),
     dig_get_element(L1, _E1, L2),
     dig_get_element(L2, E2,_),
     E2=element(_, [name=Concept|_],_),
      !,
     dig_get_shorted_name(Concept, CoreKeyword),
     true.


%get_axiom_distance(+CoreKeyword, +Axiom,  +DistanceMeasure, -Distance)

 get_axiom_distance(CoreKeyword, Axiom,  google, Distance):-
  !,
   pion_icr_set_element(Axiom, icr(_I, C, _R)),
   google_distance_element(google_distance_min, C, CoreKeyword, Distance),
    true.


get_axiom_distance(_CoreKeyword, _Axiom,  _DistanceMeasure, Distance):-
    Distance = 0.


%path_pruning(+Workingset, +CutDistance, +CoreKeyword, +DistanceMeasure, -CuttedSet),

path_pruning([],  _Distance, _CoreKeyword, _DistanceMeasure, []):-!.

    
path_pruning([E|L], Distance, CoreKeyword,  DistanceMeasure, L2):-
    get_axiom_distance(CoreKeyword, E, DistanceMeasure, Distance1),
    Distance1 > Distance,
    !,
%    format(user_error, '~w is deleted, because of the distance ~w > ~w~n', [E, Distance1, Distance]),
    path_pruning(L, Distance, CoreKeyword,  DistanceMeasure, L2).


path_pruning([E|L], Distance, CoreKeyword,  DistanceMeasure, [E|L2]):-
    !,
    path_pruning(L, Distance, CoreKeyword,  DistanceMeasure, L2).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% end of path pruning strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

query_post(L, Q, Status):-
      !,
      L1 =[element(clearKB, [], [])|L],
	dig_tell(elements(L1), answer(_, _Answer1), [connection(close)]),
 %%     format(user_error, 'Query post answer=~q~n', [Answer1]),
 %%    message_log(L1), % for log only
 %%    dig_consistency_check(S),
 %%   format(user_error,'S=~q~n', [S]),      
	dig_ask(elements([Q]), answer(_, Answer2), [connection(close)]),     
 %     format(user_error, 'Query post answer2=~q~n', [Answer2]),  
	opposite_query(Q, Q1),
	dig_ask(elements([Q1]), answer(_, Answer3), [connection(close)]), 
 %    format(user_error, 'Query post answer3=~q~n', [Answer3]),
 %    (S= false -> Status = overdetermined; answer_status(Answer2, Answer3, Status)),  
      answer_status(Answer2, Answer3, Status),
 %     format(user_error,'Status=~q~n', [Status]), 
	true.


opposite_query(Q, Q1):-
      Q = element(instance, _X, L),
	!,
      dig_get_element(L, I, L1),
      dig_get_element(L1, C, _L2),
      dig_add_elements([C], not, C1),
      Q1= element(instance, [], [I|C1]).


opposite_query(Q, Q1):-
      Q = element(subsumes, _X, L),
      !,
      dig_get_element(L, C1, L1),
      dig_get_element(L1, C2, _L2),
      C3 = element(not, [], [C1]),
      Q1= element(subsumes, [id='id'], [C3, C2]).



opposite_query(Q, Q1):-
      dig_add_elements([Q], not, [Q1]).


%explain_answer(+Status, +ID, +CurrentWorkingList, +K, +K1, -Answer)  %% version 1.2.4 added

explain_answer(Status, ID, L, K, K1, Answer):-
	(Status = accepted; Status = rejected),
      pion_setting(explain, on),
      !,
      delete_repeated_element(L, L1),
     	Answer = element(Status,[id=ID, k=K+K1], [element(because, [], L1)]).

explain_answer(Status, ID, _L, K, K1, Answer):-
     Answer = element(Status, [id=ID, k=K+K1],[]).



%explain_answer(+Status, +ID, +CurrentWorkingList, -Answer)

explain_answer(Status, ID, L, Answer):-
	(Status = accepted; Status = rejected),
      pion_setting(explain, on),
      !,
      delete_repeated_element(L, L1),
     	Answer = element(Status,[id=ID], [element(because, [], L1)]).


explain_answer(Status, ID, _L, Answer):-
     Answer = element(Status, [id=ID],[]).

delete_repeated_element([], []):-!.

delete_repeated_element([X|Y], Z):-member(X, Y),
                        !,
                        delete_repeated_element(Y, Z).

delete_repeated_element([X|Y], [X|Z]):-
                        delete_repeated_element(Y, Z).




%the +answer set has an element with 
%the pattern of the message (from the server)
%define the pattern of the incoherency message for RACER

%it may not work for other untested DIG system.

pattern_set(As, incoherent):-
%    format(user_error, 'Error pattern As:~w~n',[As]),
    E=element(error, L, _),
    member(E, As),
    dig_select_attribute(L, message, M),
    sub_string(M, _S, _L, _B, 'incoherent').  % Racer error pattern

pattern_set(As, incoherent):-
    E=element(error, L, _),
    member(E, As),
    memberchk(code='400', L),
%    memberchk(message='The empty clause was derived.', L),
    true.   % KAON2/Pellet error pattern




%The  +Element has the pattern
%define the pattern of the incoherency message for RACER

pattern(E, incoherent):-
%    format(user_error, 'Error pattern:~w~n',[E]),
    E=element(error, L, _),
    dig_select_attribute(L, message, M),
    sub_string(M, _S, _L, _B, 'incoherent').  % Racer error pattern

pattern(E, incoherent):-
    E=element(error, L, _),
    memberchk(code='400', L),
%   memberchk(message='The empty clause was derived.', L),
    true.   % KAON2/Pellet error pattern



pion_setting(X, Y):-
   dig_db_element(setting(pion, X), Y).

pion_init:-
   retractall(dig_db_element(setting(pion, _), _)), 
   assert(dig_db_element(setting(pion, selectionFunction), syntactic_relevance)),
%  assert(dig_db_element(setting(pion, selectionFunction), syntactic_concept_relevance)),
% assert(dig_db_element(setting(pion, selectionFunction), semantic_relevance_ngd)),
%  assert(dig_db_element(setting(pion, selectionFunction), google_distance_min)),
%  assert(dig_db_element(setting(pion, selectionFunction), google_distance_max)),
   assert(dig_db_element(setting(pion, googleDistanceDirectory), google_distance)),
   assert(dig_db_element(setting(pion, strategy), linear_extension)),
%  assert(dig_db_element(setting(pion, strategy), k_extension)),
   assert(dig_db_element(setting(pion, kvalue), '2')),
   assert(dig_db_element(setting(pion, kb), kb(pion,default))),
   assert(pion_log_counter(0)),
   assert(dig_db_element(setting(pion, odp), first_maximal_consistent_set)),
%  assert(dig_db_element(setting(pion, odp), path_pruning_with_google_distance)),
   assert(dig_db_element(setting(pion, odpLimit), 1000)),
   assert(pion_log_counter(odp, 0)),
   assert(dig_db_element(setting(pion, explain), off)),
   dig_newKB_init,
   load_google_distance_data,
   true.

dig_newKB_init:-
    dig_get_newKB(URI),
    (URI = nil -> format('~nIf you need no an external DL reasoner, you can ignore this error message.~n', []); true),
    assert(dig_db_element(setting(kburi, default), URI)),
    format('~n~n set newKB URI=~q.~n~n', [URI]).


pion_getting_processing(Data, A):-
    Data=[element(data, Atts, _)],
    memberchk(path=Path, Atts),
    pion_path_processing(Path, Atts, A),
    true.



pion_path_processing(Path, _Atts, _Reply):-
   Path='/',
   !,
   reply_file('index.htm').




pion_path_processing(Path, Atts, _Reply):-
   select(path=Path, Atts, Other),
   Other=[],
   !,
   sub_string(Path, 0, _Y, Z, '/'),
   sub_string(Path, 1, Z, 0, File), 
   reply_file(File).

pion_path_processing(Path, Atts, _Reply):-
   Path='/request',
   memberchk(command=getKBURI, Atts),
   memberchk(kb=KB, Atts),
   dig_db_element(setting(kburi, KB), URI),
   format('Content-type: text/plain~n~n',[]),
   format('~w~n', [URI]),
   true.




pion_path_processing(Path, Atts, Reply):-
   format('Content-type: text/html~n~n',[]),
   sformat(Reply, 'Unknown Path: ~q with parameters Atts=~q', [Path, Atts]),
   format('~w~n',[Reply]),
   true.




load_google_distance_data :-
   retractall(ngd(_,_)), 
   dig_db_element(setting(pion, googleDistanceDirectory),D),
   concat_atom([D, '/*.xml'], WildCard),
   expand_file_name(WildCard,L), 
   load_google_distance_data_list(L).


load_google_distance_data_list([]).

load_google_distance_data_list([File|L]):-
    load_data_file(File, _Attri, Data),
    format('Loading Google distance data from the file ~q~n', [File]),
    set_google_data(Data),
    !,
    load_google_distance_data_list(L).

load_data_file(File, Attri, Data):-
	open(File, read, In),
      load_structure(In, L, []),
      close(In),
      dig_get_element(L, E, _),
      E=element(_,Attri, Data),
%      retractall(file_data(File,_, _)),
%      assert(file_data(File, Attri, Data)),
      true.

set_google_data([]).

set_google_data([element(ngd, Atts,_)|L]):-
    memberchk(keyword1=K1, Atts),
    memberchk(keyword2=K2, Atts),
    memberchk(ngd=NGD, Atts),
    !,
    string_to_atom(K10, K1),
    string_to_atom(K20, K2),
%    assert(ngd(K10,K20, NGD)),
    assert(ngd(k(K10,K20), NGD)),  
    set_google_data(L).

set_google_data([_|L]):-
   set_google_data(L).

ngd(X,Y, N) :- ngd(k(X,Y), N).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% predicates for debugging and log only
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%/*
show_overdetermined_processing(E):-
          E = element(Type, _, EL),            
            E1 = element(_, [name=N], _),
            dig_get_element(EL, E1, _),
            format(user_error, 'Ignoring: axiom(~w_~w)~n~n', [Type, N]).


message_log(_L):-
      pion_log_counter(I),
	I > 500,  % limit of the log counter
      !.

message_log(L):-
      pion_log_counter(I),
      I1 is I+1,
      retract(pion_log_counter(I)),
      assert(pion_log_counter(I1)),
      concat_atom(['log/tmp_tell_log', I, '.xml'], File0),
      !,
      open(File0, write, Out, []),
	dig_xmlns(XMLNS),
	TellElements = [element(tells, XMLNS, L)],
      elements_xmltext(TellElements, file(File0)),
 	close(Out),
      true.

message_log(_L):-!.

%*/
 


			

:- pion_init, dig_server_port(Port), set_dig_server_id('SEKT PION Server 2.1.0'), http_server(dig_server,[port(Port)]).
